import torch
from torch import nn
import copy


class NN(nn.Module):
    @property
    def device(self):
        return next(self.parameters()).device


class Embedding(NN):
    def __init__(self, vocab_size, dim=512, padding_idx=0, dropout=0.1):
        super().__init__()

        self.emb = nn.Embedding(vocab_size, dim, padding_idx=padding_idx)
        nn.init.uniform_(self.emb.weight, -0.01, 0.01)
        # self.norm = nn.LayerNorm(dim)
        max_len = 1024
        pe = torch.zeros(max_len, dim)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(
            torch.arange(0, dim, 2).float() *
            (-torch.tensor(10000.0).log() / dim))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)
        self.dropout = nn.Dropout(dropout)
        self.dim = dim

    def forward(self, x, pos=0):
        x = self.emb(x) * (self.dim**0.5)
        x = x + self.pe[pos:x.size(0) + pos]
        x = self.dropout(x)
        # x = self.norm(x)
        return x


def attn(q, k, v, mask=None):
    """
    q: (ql, b, dim)
    k: (kl, b, dim)
    v: (vl, b, dim)
    mask: (b, ql, kl)
    """
    q = q.transpose(0, 1)
    k = k.transpose(0, 1)
    v = v.transpose(0, 1)
    score = (q @ k.transpose(-2, -1)) / (q.size(-1)**0.5)  # (b, ql, kl)
    if mask is not None:
        score = score.masked_fill(mask, -float('inf'))
    score = score.softmax(-1)
    out = (score @ v)
    out = out.transpose(0, 1)
    return out, score


def triu_mask(x):
    """x: (xlen, ...)"""
    xlen = x.size(0)
    mask = torch.triu(torch.ones(xlen, xlen, device=x.device), 1).bool()
    return mask  # (xlen, xlen)


def kv_mask(x, padding_idx=0, num_heads=8):
    """x: (xlen, b)"""
    xlen = x.size(0)
    mask = x.ne(padding_idx).T.unsqueeze(1)  # (b, 1, xlen)
    mask = torch.stack([mask] * num_heads, 1).view(-1, 1, xlen)
    return mask


class MultiheadAttention(NN):
    def __init__(self, dim=512, num_heads=8):
        super().__init__()

        self.q_in = nn.Linear(dim, dim, bias=False)
        self.k_in = nn.Linear(dim, dim, bias=False)
        self.v_in = nn.Linear(dim, dim, bias=False)
        self.out = nn.Linear(dim, dim, bias=False)

        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads

    def forward(self, q, k, v, mask=None):
        q_len = q.size(0)
        k_len = k.size(0)
        v_len = v.size(0)
        bsz = q.size(1)

        q = self.q_in(q)
        k = self.k_in(k)
        v = self.v_in(v)

        q = q.reshape(q_len, bsz * self.num_heads, self.head_dim)
        k = k.reshape(k_len, bsz * self.num_heads, self.head_dim)
        v = v.reshape(v_len, bsz * self.num_heads, self.head_dim)

        out, score = attn(q, k, v, mask)

        out = out.reshape(q_len, bsz, self.dim)
        out = self.out(out)
        return out


class FFN(NN):
    def __init__(self, dim=512, inner_dim=2048):
        super().__init__()

        self.l1 = nn.Linear(dim, inner_dim)
        self.relu = nn.ReLU()
        self.l2 = nn.Linear(inner_dim, dim)

    def forward(self, x):
        x = self.l1(x)
        x = self.relu(x)
        x = self.l2(x)
        return x


class EncoderLayer(NN):
    def __init__(self, dim=512, num_heads=8, dropout=0.1):
        super().__init__()

        self.self_attn = MultiheadAttention(dim, num_heads)
        self.ffn = FFN(dim, dim * 4)

        self.norm1 = nn.LayerNorm(dim)
        self.norm2 = nn.LayerNorm(dim)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        x = x + self.dropout(self.self_attn(x, x, x))
        x = self.norm1(x)
        x = x + self.dropout(self.ffn(x))
        x = self.norm2(x)
        return x


class DecoderLayer(NN):
    def __init__(self, dim=512, num_heads=8, dropout=0.1):
        super().__init__()

        self.self_attn = MultiheadAttention(dim, num_heads)
        self.ctx_attn = MultiheadAttention(dim, num_heads)
        self.ffn = FFN(dim, dim * 4)

        self.norm1 = nn.LayerNorm(dim)
        self.norm2 = nn.LayerNorm(dim)
        self.norm3 = nn.LayerNorm(dim)

        self.dropout = nn.Dropout(dropout)

    def forward(self, y, x):
        """y_step: (ylen, b, dim)
        x: (xlen, b, dim)"""
        y = y + self.dropout(self.self_attn(y, y, y, triu_mask(y)))
        y = self.norm1(y)
        y = y + self.dropout(self.ctx_attn(y, x, x))
        y = self.norm2(y)
        y = y + self.dropout(self.ffn(y))
        y = self.norm3(y)
        return y

    def forward_step(self, y_step, x, cache):
        """y_step: (1, b, dim)
        x: (xlen, b, dim)
        cache: (ylen, b, dim), in which 0 is self_attn_cache, 
                1 is ctx_attn_cache"""

        cache = y_sofar = torch.cat([cache, y_step])
        y_step = y_step + self.dropout(self.self_attn(y_step, y_sofar,
                                                      y_sofar))
        y_step = self.norm1(y_step)

        y_step = y_step + self.dropout(self.ctx_attn(y_step, x, x))
        y_step = self.norm2(y_step)

        y_step = y_step + self.dropout(self.ffn(y_step))
        y_step = self.norm3(y_step)

        return y_step, cache

    def init_cache(self, b, dim):
        """
        cache: (ylen=0, b, dim), actually the y_prev
        """
        cache = torch.empty(0, b, dim).to(self.device)
        return cache


class Encoder(NN):
    def __init__(self, layer, num_layers=6):
        super().__init__()

        self.layers = nn.ModuleList(
            [copy.deepcopy(layer) for i in range(num_layers)])

    def forward(self, x):
        for l in self.layers:
            x = l(x)
        return x


class Decoder(NN):
    def __init__(self, layer, num_layers=6):
        super().__init__()

        self.layers = nn.ModuleList(
            [copy.deepcopy(layer) for i in range(num_layers)])

    def forward(self, y, x):
        for l in self.layers:
            y = l(y, x)
        return y

    def forward_step(self, y_step, x, cache):
        """y_step: (1, b, dim)
        x: (xlen, b, dim)
        cache: (layer, ylen, b, dim), in which 0 is self_attn_cache, 
                1 is ctx_attn_cache"""
        new_cache = []
        for i, l in enumerate(self.layers):
            y_step, cache_ = l.forward_step(y_step, x, cache[i])
            new_cache.append(cache_)
        return y_step, torch.stack(new_cache)

    def init_cache(self, b, dim):
        """
        cache: (layer, ylen=0, b, dim), actually y_prev
        """
        num_layers = len(self.layers)
        cache = torch.empty(num_layers, 0, b, dim).to(self.device)
        return cache


class Transformer(NN):
    def __init__(self,
                 src_token_space_size,
                 trg_token_space_size,
                 dim=512,
                 num_layers=6,
                 num_heads=8,
                 dropout=0,
                 pad_id=0,
                 ):
        super().__init__()

        self.pad_id = pad_id
        if pad_id != 0:
            print('[error] pad_id = {}, but this Transformer model only supports pad_id = 0'.format(pad_id))
            assert False

        self.dim = dim
        self.src_emb = Embedding(src_token_space_size,
                                 dim=dim,
                                 padding_idx=self.pad_id,
                                 dropout=dropout)
        self.trg_emb = Embedding(trg_token_space_size,
                                 dim=dim,
                                 padding_idx=self.pad_id,
                                 dropout=dropout)
        self.trg_emb.emb.weight = self.src_emb.emb.weight
        self.proj = nn.Linear(dim, trg_token_space_size, bias=False)
        self.proj.weight = self.trg_emb.emb.weight

        encoder_layer = EncoderLayer(dim, num_heads, dropout)
        self.encoder = Encoder(encoder_layer, num_layers)
        decoder_layer = DecoderLayer(dim, num_heads, dropout)
        self.decoder = Decoder(decoder_layer, num_layers)

        #hack: name placeholders for legacy code; not used now
        self.src_tok = None #src_tok
        self.trg_tok = None #trg_tok

    def forward_batch(self, x, y):
        # used for training
        x = self.src_emb(x)
        x = self.encoder(x)
        y = self.trg_emb(y)
        y = self.decoder(y, x)
        y = self.proj(y)
        return y
