import numpy as np
import matplotlib.pyplot as plt


class RV1d(object):
    def __init__(self, name='null', save_data=False):
        self.name = name
        self.sum = 0
        self.squared_sum = 0
        self.cnt = 0
        self.sample = [] if save_data else None

    def append(self, x):
        self.sum += x
        self.squared_sum += x**2
        self.cnt += 1
        if self.sample is not None: self.sample.append(x)

    def merge(self, another_rv):
        self.sum += another_rv.sum
        self.squared_sum += another_rv.squared_sum
        self.cnt += another_rv.cnt
        if self.sample is not None:
            self.sample = None if another_rv.sample is None else self.sample + another_rv.sample

    #@property
    def size(self):
        return self.cnt

    #@property
    def mean(self):
        return float('nan') if self.cnt < 1 else self.sum / self.cnt

    #@property
    def var(self):
        return float('nan') if self.cnt < 2 else (self.squared_sum - self.sum**2 / self.cnt) / (self.cnt-1)

    #@property
    def stdev(self):
        var = self.var()
        assert var >= 0  # var can be negative when S^2 is strictly 0, due to float arithmetic
        return np.sqrt(var)

    def mean_ci(self, cl=90, force=True):
        """
        Provide an approximate confidence interval of the population mean based on CLT:
        The interval has a prior-probability of 'cl' (by default 90%) to capture the true population mean under apriori
        assumption that the sample size has been "large enough" to make the sample mean normally distributed (and to
        make the sample stdev converges as well).
        cl: confidence level. currently only support cl=90, 95, 98, 99
        force: by default force=True, in which case will output an actual estimate whenever possible (i.e. as long as
                sample size >= 2); when force=False, will output (-inf,+inf) if the sample size is judged too small.
        """
        if self.cnt < 2: return float('-inf'), float('+inf')

        if force is False:
            # TODO: come up with better method to judge whether the sample size is large enough for quality output,
            #  e.g. based on normality test
            if self.cnt < 50: return float('-inf'), float('+inf')

        sample_size = self.cnt
        sample_mean = self.mean()
        sample_stdev = self.stdev()
        quantile = {90:1.645, 95:1.960, 98:2.326, 99:2.576}  # two-sided confidence level
        error = sample_stdev * quantile[cl] / np.sqrt(sample_size)
        interval = (sample_mean - error, sample_mean + error)
        return interval

    def __str__(self):
        name = self.name
        mean = self.mean()
        mean_interval = self.mean_interval()
        dev = self.dev()
        size = self.size()
        return '%s\tmean= %.2f (%.2e , %.2e)\tstd= %.2f\tn= %d' % (name, mean, *mean_interval, dev, size)

    def plot(self, blocking=True):
        if self.sample is not None:
            self._plot(blocking)

    def _plot(self, blocking=True):
        data = self.sample

        if not blocking:
            plt.ion()
            #plt.draw()
            #plt.pause(0.1)

        fig, ax = plt.subplots()
        plt.hist(data, bins='auto', facecolor='blue', alpha=0.5)
        fig.suptitle(self.name)
        ax.grid()
        #plt.subplots_adjust(left=0.15)

        if not blocking:
            plt.draw()
            plt.pause(0.1)
        else:
            plt.show(block=True)

    # back-compatible interface
    #@property
    def dev(self):
        return self.stdev()

    #@property
    def mean_interval(self):
        return self.mean_ci(cl=90)


if __name__ == "__main__":
    X = RV1d('X', save_data=True)
    for i in range(10):
        X.append(i)
    print(X)
    print(X.mean())
    print(X.mean_interval())
    print(X.mean_ci(cl=90))
    print(X.mean_ci(force=False))
    print(X.mean_ci(cl=95))
    print(X.dev())
    print(X.stdev())
    print(X.var())
    X.plot(blocking=False)
    X.plot()

    Y = RV1d('Y', save_data=True)
    for i in range(10,20):
        Y.append(i)
    print(Y)
    Y.merge(X)
    print(Y)
    Y.plot()  # the empirical distribution is not uniform because the bin width is 20/6=3.3

    Y = RV1d('Y without data')
    for i in range(10, 20):
        Y.append(i)
    print(Y)
    X.merge(Y)  # X should lost its original data because Y's data has been lost
    X.plot()  # should see nothing
    '''
    X	mean= 4.50 (2.93e+00 , 6.07e+00)	std= 3.03	n= 10
    4.5
    (2.925032407740824, 6.074967592259176)
    (2.925032407740824, 6.074967592259176)
    (-inf, inf)
    (2.6234428687975773, 6.376557131202423)
    3.0276503540974917
    3.0276503540974917
    9.166666666666666
    Y	mean= 14.50 (1.29e+01 , 1.61e+01)	std= 3.03	n= 10
    Y	mean= 9.50 (7.32e+00 , 1.17e+01)	std= 5.92	n= 20
    Y without data	mean= 14.50 (1.29e+01 , 1.61e+01)	std= 3.03	n= 10
    '''


